//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

// SAT3 MODULE EXAMPLE - SIMPLE LINEAR INTERPOLATION ROUTINE
// x1 y1 x2 y2 x -> y=y1+(y2-y1)/(x2-x1)*(x-x1)

#include <hpgcc3.h>


// THIS STATEMENT OPTIMIZES THE PROGRAM TO EXECUTE FROM STACK
// USING THE CALCULATOR IN SLOW MODE, DISABLING INITIALIZATION OF
// ANY CONSOLE INPUT/OUTPUT AND RESERVING MORE MEMORY FOR
// SATURN OBJECTS CREATION. SEE THE DOCUMENTATION ON CONFIG MODULE.

// IT IS DISABLED FOR DEBUGGING, TO GET CONSOLE OUTPUT.
//ARMCODE_CONFIG_STACK_SLOW

void numeric_to_xreal(SAT_OBJECT obj,XREAL *x)
{
	printf("Converting from... ");
	// CONVERT ANY NUMERIC OBJECT obj TO XREAL FORMAT
	if(sat3_isreal(obj)) { sat3_convertreal_xreal(obj,x); printf("real\n"); }
	if(sat3_isbint(obj)) { xr_int2xr(sat3_convertbint_int(obj),x); printf("bint\n"); }
	if(sat3_iszint(obj)) { sat3_convertzint_xreal(obj,x); printf("zint\n"); }
	if(sat3_ishxs(obj)) { xr_llong2xr(sat3_converthxs_llong(obj),x); printf("hxs\n"); }
}
SAT_OBJECT xreal_to_objtype(SAT_OBJECT obj,XREAL *x)
{
	// CONVERT AN XREAL NUMBER x INTO A NEW OBJECT OF SAME TYPE AS OBJ
	if(sat3_isreal(obj)) return sat3_createreal_xreal(ALLOC_TEMPOB,x);
	if(sat3_isbint(obj)) return sat3_createbint_int(ALLOC_TEMPOB,xr_xr2int(x));
	if(sat3_iszint(obj)) return sat3_createzint_xreal(ALLOC_TEMPOB,x);
	if(sat3_ishxs(obj)) return sat3_createhxs_llong(ALLOC_TEMPOB,xr_xr2llong(x));
	return 0;
}


int main()
{
	// USE THE XREAL TYPE, WHICH CAN SAFELY CONTAIN CALCULATOR REALS WITHOUT LOSS OF PRECISION
	// DECNUMBER COULD ALSO BE USED HERE
	XREAL x1,y1,x2,y2,x,y;
	SAT_OBJECT x1s,y1s,x2s,y2s,xs,ys;


	printf("Starting stack example...\n");

	// THE CONSTANT SAT_NUMERIC IS A SPECIAL WAY TO DESCRIBE CALCULATOR NUMBERS
	// IN ANY FORMAT: BINT (SYSRPL INTEGERS), HXS (USERRPL INTEGERS), ZINT (INTEGERS), AND REAL FORMATS
	// CHECK FOR 5 NUMERIC ARGUMENTS
	sat3_checkargserror(5,SAT_NUMERIC,SAT_NUMERIC,SAT_NUMERIC,SAT_NUMERIC,SAT_NUMERIC);

	printf("Arguments were OK at this point.\n");

	// POP VALUES FROM THE STACK
	// THE USER WILL PUSH IN ORDER x1 y1 x2 y2 x SO WE POP IN REVERSE ORDER
	xs=sat3_pop();
	y2s=sat3_pop();
	x2s=sat3_pop();
	y1s=sat3_pop();
	x1s=sat3_pop();

	// CONVERT ALL OUR ARGUMENTS TO XREAL FORMAT TO OPERATE
	// WE COULD CONVERT TO double FOR SPEED IF WE DON'T CARE FOR POSSIBLE
	// EXPONENT OVERFLOW (CALCULATOR REALS CAN HAVE +/-499, double CAN ONLY REACH +/-308)

	numeric_to_xreal(x1s,&x1);
	numeric_to_xreal(y1s,&y1);
	numeric_to_xreal(x2s,&x2);
	numeric_to_xreal(y2s,&y2);
	numeric_to_xreal(xs,&x);

	// PRINT ALL ARGUMENTS FOR DEBUG PURPOSES
	// IMPORTANT: %llf IS A NON-STANDARD HPGCC3-ONLY EXTENSION TO printf
	//            IT PRINTS XREAL NUMBERS PASSED AS POINTERS
	//            THE COMPILER WILL ALWAYS EMIT A WARNING, SIMPLY IGNORE IT
	//            SINCE IT IS A VALID FORMAT IN HPGCC3.
	printf("x1=%llf\n",&x1);
	printf("y1=%llf\n",&y1);
	printf("x2=%llf\n",&x2);
	printf("y2=%llf\n",&y2);
	printf("x=%llf\n",&x);

	// NOW LET'S DO THE MATH

	XREAL dx,dy,dx2;

	// dx=x2-x1;
	xr_sub(&x2,&x1,&dx);

	// dy=y2-y1;
	xr_sub(&y2,&y1,&dy);

	// dx2=x-x1;
	xr_sub(&x,&x1,&dx2);

	// dx2=(y2-y1)*(x-x1)/(x2-x1)
	xr_mul(&dy,&dx2,&dx2);
	xr_div(&dx2,&dx,&dx2);

	// y=y1+dx2=y1+(y2-y1)*(x-x1)/(x2-x1)
	xr_add(&y1,&dx2,&y);

	// PRINT INTERPOLATED VALUE FOR y
	printf("\ny=%llf\n",&y);

	// CONVERT BACK TO SAME CALCULATOR FORMAT AS x
	// IF x IS INTEGER THE RESULT MIGHT BE TRUNCATED TO BECOME AN INTEGER
	ys=xreal_to_objtype(xs,&y);


	// CHECK FOR ERROR, ASSUME INSUFFICIENT MEMORY SINCE IT'S MOST LIKELY CAUSE FOR FAILURE
	if(!ys) sat3_exiterror(SATERR_INSUFFMEMORY);

	printf("Pushing y on stack...\n");
	// PUSH THE VALUE ON THE STACK
	sat3_push(ys);

	printf("Exit\n");
	getchar();

	return 0;

}
